from highway_star.scrapping.wikipedia_scraper import scrap_wikipedia_structure_with_content
from highway_star.preprocessing.biography_preprocessor import sent_to_words
from highway_star.preprocessing.biography_preprocessor import remove_stop_words_from_biographies
from highway_star.visualizing.visualizer import give_sankey_data_from_prefixspan
from highway_star.visualizing.visualizer import sankey_diagram_with_prefixspan_output
import pandas as pd


def run(root_category: str, lang: str, custom_stop_words: list, use_lemmatization: bool = False,
        allowed_postags: list = ['NOUN', 'VERB'], prefixspan_minlen: int = 10, prefixspan_topk: int = 50,
        js_filename: str = "sankey", html_filename: str = "index", title: str = "sankey"):
    content = scrap_wikipedia_structure_with_content(root_category=root_category, lang=lang)
    dataframe_with_biographies = pd.DataFrame.from_dict(content)
    dataframe_with_biographies["biographies_tokenized"] = sent_to_words(
        biographies_column=dataframe_with_biographies["content"])
    dataframe_with_biographies["biographies_cleaned"] = remove_stop_words_from_biographies(
        biographies_column=dataframe_with_biographies["biographies_tokenized"], custom_stop_words=custom_stop_words,
        use_lemmatization=use_lemmatization, allowed_postags=allowed_postags)
    sankey_data_from_prefixspan = give_sankey_data_from_prefixspan(
        data_tokenized=dataframe_with_biographies["biographies_cleaned"], prefixspan_minlen=prefixspan_minlen,
        prefixspan_topk=prefixspan_topk)
    sankey_diagram_with_prefixspan_output(sankey_data_from_prefixspan=sankey_data_from_prefixspan,
                                          js_filename=js_filename, html_filename=html_filename, title=title)



from gensim.utils import simple_preprocess
from nltk import RegexpTokenizer
from nltk.corpus import stopwords
import spacy
from typing import List
from time import time
import pandas
from rich.progress import track
import nltk


def sent_to_words(biographies_column: pandas.core.series.Series) -> list:
    t = time()
    biographies_column = biographies_column.astype('str')
    tokenizer = RegexpTokenizer(r'\w+')
    biographies_tokenized = list(biographies_column.apply(tokenizer.tokenize).values)
    print('Time to tokenize everything: {} mins'.format(round((time() - t) / 60, 2)))
    return biographies_tokenized


def lemmatization(tokenized_biographies: list, allowed_postags=['NOUN', 'VERB'], lang: str = "fr") -> list:
    texts_out = []
    for sent in track(tokenized_biographies, description="lemmatize biographies"):
        nlp = ""
        if lang == "fr":
            nlp = spacy.load('fr_core_news_sm')
        elif lang == "eng":
            nlp = spacy.load('en_core_web_sm')
        doc = nlp(" ".join(sent))
        texts_out.append([token.lemma_ for token in doc if token.pos_ in allowed_postags])
    return texts_out


def remove_stopwords(tokenized_biographies: list, stop_words) -> list:
    return [[word for word in simple_preprocess(str(doc)) if word not in stop_words] for doc in
            track(tokenized_biographies, description="removing stop words")]


def remove_stop_words_from_biographies(
        biographies_column: pandas.core.series.Series,
        custom_stop_words: List[str] = None,
        use_lemmatization=False,
        allowed_postags=['NOUN', 'VERB'],
        lang: str = "fr") -> list:
    tokenized_biographies: list = sent_to_words(biographies_column=biographies_column)
    nltk.download("stopwords")
    stop_words: List[str] = [""]
    if lang == "fr":
        stop_words: List[str] = stopwords.words('french')
    if lang == "eng":
        stop_words: List[str] = stopwords.words('english')
    if custom_stop_words:
        stop_words.extend(custom_stop_words)
    if use_lemmatization:
        tokenized_biographies = lemmatization(tokenized_biographies, allowed_postags=allowed_postags, lang=lang)
    return remove_stopwords(tokenized_biographies=tokenized_biographies, stop_words=stop_words)



from bs4 import BeautifulSoup, ResultSet
import requests
import re
from rich.progress import track
from typing import List
from SPARQLWrapper import SPARQLWrapper, JSON


def get_subcategories_names_from_category(category: str, lang: str) -> List[str]:
    subcategories_names = []
    category_root = ""
    if lang == "fr":
        category_root = 'https://fr.wikipedia.org/wiki/Catégorie:'
    elif lang == "eng":
        category_root = 'https://en.wikipedia.org/wiki/Category:'
    soup: BeautifulSoup = BeautifulSoup(requests.get(category_root + category).content,
                                        "html.parser")
    category: BeautifulSoup = soup.findAll("div", {"class": "mw-category-generated"})[0]
    try:
        sub_categories: BeautifulSoup = category.findAll("div", {"id": "mw-subcategories"})[0].findAll(
            "div", {"class": "mw-content-ltr"})[0].findAll("a")
        for sub_category in sub_categories:
            subcategories_names.append(sub_category.get_text().replace(" ", "_"))
    except IndexError:
        pass
    return subcategories_names


def get_pages_from_category(category: str, lang: str) -> dict:
    pages_names = []
    pages_links = []
    subcategory = []
    category_root = ""
    wikipedia_root = ""
    if lang == "fr":
        category_root = 'https://fr.wikipedia.org/wiki/Catégorie:'
        wikipedia_root = "https://fr.wikipedia.org"
    elif lang == "eng":
        category_root = 'https://en.wikipedia.org/wiki/Category:'
        wikipedia_root = "https://en.wikipedia.org/"
    soup: BeautifulSoup = BeautifulSoup(requests.get(category_root + category)
                                        .content, "html.parser")
    try:
        category_generated: BeautifulSoup = soup.findAll("div", {"class": "mw-category-generated"})[0]
        pages: BeautifulSoup = category_generated.findAll("div", {"id": "mw-pages"})[0].findAll(
            "div", {"class": "mw-content-ltr"})[0].findAll("a")
        for page in pages:
            pages_links.append(wikipedia_root + page["href"])
            pages_names.append(page.get_text().replace(" ", "_"))
            subcategory.append(category)
    except IndexError:
        pass
    pages_dictionary: dict = {"pages_links": pages_links, "pages_names": pages_names, "subcategory": subcategory}
    return pages_dictionary


def scrap_wikipedia_structure(subcategories: List[str], lang: str,
                              pages_dictionary: dict = dict(pages_links=[], pages_names=[], subcategory=[])) -> dict:
    if subcategories:
        new_subcategories: List[str] = []
        for names in track(subcategories, description="Crawling " + str(len(subcategories)) + " subcategories"):
            new_subcategories.extend(get_subcategories_names_from_category(names, lang))
            pages: dict = get_pages_from_category(names, lang)
            pages_dictionary["pages_links"].extend(pages["pages_links"])
            pages_dictionary["pages_names"].extend(pages["pages_names"])
            pages_dictionary["subcategory"].extend(pages["subcategory"])
        return scrap_wikipedia_structure(subcategories=new_subcategories, pages_dictionary=pages_dictionary, lang=lang)
    else:
        return pages_dictionary


def get_specific_content_from_page(page_content: BeautifulSoup, start_tag: str, end_tag: str) -> str:
    biography: str = ""
    try:
        paragraphs_in_biography: ResultSet = BeautifulSoup(
            str(page_content).split(start_tag)[1].split(end_tag)[0], "html.parser").findAll("p")
        for paragraphs in paragraphs_in_biography:
            biography += re.sub(r'\[.*\]', '',
                                paragraphs
                                .get_text()
                                .replace("«", "« ")
                                .replace("»", " »")
                                .replace("\xa0", "")
                                .replace("\n", "")
                                .replace(".", " . ")
                                .replace(",", " , ")
                                .replace(";", " ; ")
                                .replace(":", " : "))
    except IndexError:
        biography: str = ""
    return biography


def get_content_from_all_pages(all_pages: dict, start_tag: List[str], end_tag: str) -> List[str]:
    all_content: List[str] = []
    for link in track(all_pages["pages_links"], description="Getting all content from pages"):
        page_content: BeautifulSoup = BeautifulSoup(requests.get(link).content, "html.parser")
        content = ""
        for i in range(len(start_tag)):
            content += " " + get_specific_content_from_page(page_content, start_tag[i], end_tag)
        if content == "     " or content == " ":
            content = "no_matching_content"
        all_content.append(content)
    return all_content


def scrap_wikipedia_structure_with_content(root_category: str, lang: str = "fr") -> dict:
    start_tag: List[str] = []
    end_tag: str = ""
    if lang == "fr":
        start_tag = ['<span class="mw-headline" id="Biographie">Biographie</span>']
        end_tag = '<h2>'
    elif lang == "eng":
        start_tag = ['<span class="mw-headline" id="Early_life">Early life</span>',
                     '<span class="mw-headline" id="Early_life_and_career">Early life and career</span>',
                     '<span class="mw-headline" id="Career">Career</span>',
                     '<span class="mw-headline" id="Death">Death</span>',
                     '<span class="mw-headline" id="Later_life_and_death">Later life and death</span>']
        end_tag = '<h2>'
    all_pages: dict = scrap_wikipedia_structure(subcategories=[root_category], lang=lang)
    content: List[str] = get_content_from_all_pages(all_pages=all_pages, start_tag=start_tag, end_tag=end_tag)
    all_pages["content"]: dict = content
    return all_pages


def query_resource_from_page_name(resource: str, all_names: List[str]) -> List[str]:
    try:
        all_names: List[str] = list(set(all_names))
        resource_in_all_pages: List = []
        for name in track(range(len(all_names)), description="Querying " + resource):
            sparql: SPARQLWrapper = SPARQLWrapper("http://dbpedia.org/sparql")
            if resource == "birthdate":
                sparql.setQuery("""PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                SELECT ?birthdate
                WHERE {
                <http://dbpedia.org/resource/""" + all_names[name] + """> dbo:birthDate ?birthdate.}
                """)
            elif resource == "birthplace":
                sparql.setQuery("""PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                SELECT ?birthplace
                WHERE {
                <http://dbpedia.org/resource/""" + all_names[name] + """> dbo:birthPlace ?birthplace.}
                """)
            elif resource == "deathdate":
                sparql.setQuery("""PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                SELECT ?deathdate
                WHERE {
                <http://dbpedia.org/resource/""" + all_names[name] + """> dbo:deathDate ?deathdate.}
                """)
            elif resource == "deathplace":
                sparql.setQuery("""PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                SELECT ?deathplace
                WHERE {
                <http://dbpedia.org/resource/""" + all_names[name] + """> dbo:deathPlace ?deathplace.}
                """)
            sparql.setReturnFormat(JSON)
            results = sparql.query().convert()
            if results["results"]["bindings"]:
                results = results["results"]["bindings"][0]
                resource_in_all_pages.append(results[resource]["value"])
            else:
                resource_in_all_pages.append(None)
        return resource_in_all_pages
    except KeyError:
        print(resource, "is not a valid ressource name, please enter another one.")



from prefixspan import PrefixSpan
from time import time
import pandas
from rich.progress import track


def give_sankey_data_from_prefixspan(data_tokenized: pandas.core.series.Series,
                                     prefixspan_minlen=10,
                                     prefixspan_topk=50) -> object:
    t = time()
    print("Executing Prefixspan...")
    prefixspan_initialized = PrefixSpan(data_tokenized)
    prefixspan_initialized.minlen = prefixspan_minlen
    top_50 = prefixspan_initialized.topk(prefixspan_topk)
    print('Time to execute prefixspan: {} min'.format(round((time() - t) / 60, 2)))

    all_combinations = []
    for array in track(top_50, description="Process sankey data from prefixspan"):
        for i in range(len(array[1])):
            try:
                number_of_elements = 0
                words = []
                if array[1][i] != array[1][i + 1]:
                    words = [array[1][i], array[1][i + 1]]
                    for j in range(len(top_50)):
                        ponderer = top_50[j][0]
                        for k in range(len(top_50[j][1])):
                            if top_50[j][1][k] == words[0] and top_50[j][1][k + 1] == words[1]:
                                number_of_elements += ponderer
                    words.append(number_of_elements)
                    all_combinations.append(words)
            except IndexError:
                pass

    return pandas.DataFrame(all_combinations).drop_duplicates().values.tolist()


def write_js(sankey_data: list, title: str) -> str:
    beginning_js = """Highcharts.chart('container', {

        title: {
            text: '"""+title+"""'
        },
        accessibility: {
            point: {
                valueDescriptionFormat: '{index}. {point.from} to {point.to}, {point.weight}.'
            }
        },
        series: [{
            keys: ['from', 'to', 'weight'],
            data: """
    end_js = """,
            type: 'sankey',
            name: 'Sankey demo series'
        }]

    });"""
    return beginning_js + str(sankey_data) + end_js


def write_html(title: str, js_filename: str) -> str:
    return """<!doctype html><html lang="en"><head><meta charset="utf-8"><title>""" + title + """</title>
  <meta name="description" content="The HTML5 Herald">
  <meta name="author" content="SitePoint">
  <script src="https://code.highcharts.com/highcharts.js"></script>
  <script src="https://code.highcharts.com/modules/sankey.js"></script>
  <script src="https://code.highcharts.com/modules/exporting.js"></script>
  <script src="https://code.highcharts.com/modules/export-data.js"></script>
  <script src="https://code.highcharts.com/modules/accessibility.js"></script> 
</head>
<body>
<figure class="highcharts-figure">
  <div id="container"><script src='""" + js_filename + """.js'></script></div>
</figure>
</body>
</html>"""


def sankey_diagram_with_prefixspan_output(sankey_data_from_prefixspan: list, js_filename="data", html_filename="page",
                                          title=None):
    if title is None:
        title = html_filename
    text_file = open(js_filename+".js", "w")
    text_file.write(write_js(sankey_data_from_prefixspan, title=title))
    text_file.close()
    print("js file written successfully !")
    text_file = open(html_filename + ".html", "w")
    text_file.write(write_html(html_filename, js_filename))
    text_file.close()
    print("html file written successfully !")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = []
with open('requirements.txt', 'r') as fh:
    for line in fh:
        requirements.append(line.strip())

setuptools.setup(
    name="highway_star",
    version="0.0.8.2",
    author="Mathéo Daly",
    author_email="matheodaly.md@gmail.com",
    description="A library to scrap content from wikipedia categories",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/matheo-daly/highway_star",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    python_requires=">=3.6",
)
